module StudentsPerformanceInExams exposing (main)

-- elm stuff
import Browser
import Html exposing (Html, text, pre, div)
import Http
import Dict
import Maybe.Extra
-- Decode and Date, Time stuff
import Html.Parser
import Parser
import Csv
import Csv.Decode
import Date
import Time exposing (Month(..))
-- Svg stuff
import TypedSvg exposing (circle, g, rect, style, svg, text_)
import TypedSvg.Attributes exposing (class, fontFamily, fontSize, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, height, r, width, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Transform(..))
-- visualisation
import Scale exposing (ContinuousScale)
import Scale.Color
import Statistics
-- other
import Color

-- MAIN
main =
  Browser.element
    { init = init
    , update = update
    , subscriptions = subscriptions
    , view = view
    }

-- MODEL
type alias Model = { httpState : HttpState
                   , fullTexts : List {description: String, data: String}
                   , data : List Data
                   , error : String
                   }

type HttpState
  = Failure
  | Loading Int
  | Success

type Arrangement
  = Line_By_Line
  | Forth_Loop

type alias Data = {description: String, data: List ( String, Maybe Float )}

filesToLoad =
  [ "StudentsPerformance.csv"
  ]

init : () -> (Model, Cmd Msg)
init _ =
  ( Model (Loading <| List.length filesToLoad) [] [] ""
  , Cmd.batch <|
      List.map (\file ->
                  Http.get
                    { url = file
                    , expect = Http.expectString <| GotText file
                    }
               ) filesToLoad
  )

-- UPDATE
type Msg
  = GotText String (Result Http.Error String)

update : Msg -> Model -> ( Model, Cmd Msg)
update msg model = (updateModel msg model, Cmd.none)

updateModel : Msg -> Model -> Model
updateModel msg model =
  case msg of
    GotText file result ->
      case result of
        Ok fullText ->
          case model.httpState of
            Failure           -> {model | httpState = Failure}
            Loading filesLeft ->
                if filesLeft-1 == 0 then
                    let newFullTexts = {description = file, data = fullText}::model.fullTexts
                        checkedFullTexts = List.map checkInputFormat newFullTexts
                    in
                    { model | httpState = Success
                            , fullTexts = checkedFullTexts
                            , data      = prepareData <|
                                            List.map
                                              (\data -> Data data.description <| csvString_to_data data.data)
                                              checkedFullTexts
                    }
                else{ model | httpState = Loading (filesLeft-1)
                            , fullTexts = {description = file, data = fullText}::model.fullTexts
                    }
            Success           -> {model | httpState = Failure}
        Err e -> { model | httpState = Failure
                         , error = httpErrorToString e
                 }

-- SUBSCRIPTIONS
subscriptions : Model -> Sub Msg
subscriptions model = Sub.none

-- VIEW
view : Model -> Html Msg
view model =
  case model.httpState of
    Failure ->
      text model.error

    Loading index->
      text ("Loading... "++(String.fromInt index)++" left")

    Success ->
      text <| Maybe.withDefault "Nothing here" <| List.head <| List.map .data model.fullTexts

    {-
      let rawlevel : Int -> List (Int,Int)
          rawlevel years = [(years,1),(1,12),(4,1),(1,5),(1,1)]

          fstData     = (list_get model.data 0).data
          numberDays  = List.length fstData
          numberYears = let years = List.map (\date -> case date of
                                                         Ok aDate -> Date.weekYear aDate
                                                         Err _    -> 2000
                                    ) <| List.map Date.fromIsoString <| List.map Tuple.first fstData
                        in (Maybe.withDefault 0 <| List.maximum years) - (Maybe.withDefault 0 <| List.minimum years)
          rawlevel_uncorrected = rawlevel numberYears
          numberSlots_uncorrected = List.foldl (*) 1 <| List.map (\(x,y)-> x*y) <| rawlevel_uncorrected
          diffDays = toFloat <| numberDays-numberSlots_uncorrected
          daysPerYear = toFloat <| List.foldl (*) 1 <| List.map (\(x,y)-> x*y) <| List.drop 1 <| rawlevel numberYears
          correctedYears = if numberSlots_uncorrected > numberDays then numberYears
                           else numberYears + (ceiling <| diffDays / daysPerYear)

          rawlevel_corrected = rawlevel correctedYears

          numberSlots = List.foldl (*) 1 <| List.map (\(x,y)-> x*y) rawlevel_corrected

          level = augmentLevel rawlevel_corrected
          drawList_Line_by_Line = drawArrangement Line_By_Line level
          drawList_Forth_Loop   = drawArrangement Forth_Loop   level
          w = 50
          h = 20
      in
      div []
        [ Html.div [] [Html.text <| "Anzahl Tage pro Zeitreihe:"++(String.fromInt numberDays)]
        , Html.div [] [Html.text <| "Anzahl Slots pro Zeitreihe: "++(String.fromInt numberSlots)]
        , Html.table []
            <| List.append
            [ Html.tr []
                [ Html.td [] []
                , Html.td [] [Html.text "Line by Line Arrangement"]
                , Html.td [] [Html.text "Forth-Loop Arrangement"]
                ]
            ]
            <| List.map (\zeitreihe ->
                 Html.tr []
                   [ Html.td [] [Html.text zeitreihe.description]
                   , Html.td [] [viewData (w,h) drawList_Line_by_Line zeitreihe]
                   , Html.td [] [viewData (w,h) drawList_Forth_Loop   zeitreihe]
                   ]
               ) model.data
        ]-}

-- viewHelper
viewArrangement : Arrangement -> List ( Int, Int ) -> Html Msg
viewArrangement arrangement drawList =
    Html.div []
        [ Html.text <| " Arrangement:"
        , Html.table []
            <| List.map (\line ->
                            Html.tr [] <| List.map (\n -> Html.td [] [Html.text <| String.fromInt n]) line
                        ) (coordsListToListOfLines drawList)
        ]

viewDataWithDescription : (Int,Int) -> List ( Int, Int ) -> Data -> Html Msg
viewDataWithDescription (w,h) drawList zeitreihe =
    Html.div []
     [ Html.h4 [] [Html.text zeitreihe.description]
     , viewData (w,h) drawList zeitreihe
     ]

viewData : (Int,Int) -> List ( Int, Int ) -> Data -> Html Msg
viewData (w,h) drawList zeitreihe =
    Html.div [] [svg [ viewBox 0 0 (toFloat w) (toFloat h)
        , TypedSvg.Attributes.width <| TypedSvg.Types.Percent 100
        , TypedSvg.Attributes.height <| TypedSvg.Types.Percent 100
        ][ drawDimension (w,h) drawList zeitreihe.data ]]

-- Decode csv-String
checkInputFormat : {description: String, data: String} -> {description: String, data: String}
checkInputFormat input =
  if String.left 15 input.data == "<!DOCTYPE HTML>" then
    let -- Error handling for parsing
        deadEndsToString : List Parser.DeadEnd -> String -- sadly Parser.deadEndsToString always just prints "TODO: deadEndsToString"
        deadEndsToString deadEnds = "[ "++(List.foldl (++) " ]" <| List.map deadEndToString deadEnds)

        deadEndToString : Parser.DeadEnd -> String
        deadEndToString {row,col,problem} = "row = "++(String.fromInt row)++", col = "++(String.fromInt col)++", problem = "++(problemToString problem)

        problemToString : Parser.Problem -> String
        problemToString problem =
          case problem of
            Parser.Expecting string        -> "Expecting \""++string++"\""
            Parser.ExpectingInt            -> "ExpectingInt"
            Parser.ExpectingHex            -> "ExpectingHex"
            Parser.ExpectingOctal          -> "ExpectingOctal"
            Parser.ExpectingBinary         -> "ExpectingBinary"
            Parser.ExpectingFloat          -> "ExpectingFloat"
            Parser.ExpectingNumber         -> "ExpectingNumber"
            Parser.ExpectingVariable       -> "ExpectingVariable"
            Parser.ExpectingSymbol string  -> "ExpectingSymbol \""++string++"\""
            Parser.ExpectingKeyword string -> "ExpectingKeyword \""++string++"\""
            Parser.ExpectingEnd            -> "ExpectingEnd"
            Parser.UnexpectedChar          -> "UnexpectedChar"
            Parser.Problem string          -> "Problem \""++string++"\""
            Parser.BadRepeat               -> "BadRepeat"
        -- get the csv string from html document
        data = String.dropLeft 15 input.data -- the parser can't parse the <!DOCTYPE HTML>-part of Html
        htmlDataResult = Html.Parser.run data
        dataInCorrectFormat =
          case htmlDataResult of
            Ok htmlData ->
              let allTexts = findallTexts htmlData

                  findallTexts : List Html.Parser.Node -> List String
                  findallTexts nodes = List.concat <| List.map2 findallTextsHelper nodes <| List.repeat (List.length nodes) []
                  findallTextsHelper : Html.Parser.Node -> List String -> List String
                  findallTextsHelper outerNode texts =
                    case outerNode of
                      Html.Parser.Text string                     -> string::texts
                      Html.Parser.Element string attrs childNodes ->
                        List.concat
                          [ List.concat <| List.map2 findallTextsHelper childNodes <| List.repeat (List.length childNodes) []
                          , texts
                          ]
                      Html.Parser.Comment string                  -> []
              in -- The longest String should be the csv one
                 Maybe.withDefault "There is no text in this html-file." <| list_Last <| List.sortBy String.length allTexts
            Err errorList -> "Input was Html-file, but Error while Parsing it: "++(deadEndsToString errorList)++"\t\r\n"
                             ++"Html-file:\t\r\n"++data

    in {input | data = dataInCorrectFormat }
  else input

csvString_to_data : String -> List ( String, Maybe Float )
csvString_to_data csvRaw =
    Csv.parse csvRaw
        |> Csv.Decode.decodeCsv decodeStockDay
        |> Result.toMaybe
        |> Maybe.withDefault []

decodeStockDay : Csv.Decode.Decoder (( String, Maybe Float ) -> a) a
decodeStockDay =
    Csv.Decode.map (\a b -> ( a, Just b ))
        (Csv.Decode.field "Date" Ok
            |> Csv.Decode.andMap
                (Csv.Decode.field "Open"
                    (String.toFloat >> Result.fromMaybe "error parsing string")
                )
        )

prepareData : List Data -> List Data
prepareData rawData =
  let startDate = Date.fromCalendarDate 1980 Dec 23
      untilDate = Date.fromCalendarDate 2011 Jun 9
      allDates  = Dict.fromList <|
                    List.map
                      (\date -> (Date.toIsoString date, Nothing))
                      <| Date.range Date.Day 1 startDate untilDate

      rawDictData : List ( Dict.Dict String (Maybe Float) )
      rawDictData = List.map Dict.fromList <| List.map .data rawData

      resultDictData : List ( Dict.Dict String (Maybe Float) )
      resultDictData = List.map (\rawDict -> Dict.union rawDict allDates) rawDictData

      resultDataList = List.map Dict.toList resultDictData

      resultWithoutWeekends = List.map (\data -> List.filter (\(dateString,mayFloat) ->
                                  case Date.fromIsoString dateString of
                                   Ok date ->  if List.member (Date.weekdayNumber date) [6,7]
                                               then False else True
                                   Err _   -> True
                              ) data) resultDataList

  in List.map2 (\raw result -> {raw | data = result}) rawData resultWithoutWeekends

dataToString : List ( String, Maybe Float ) -> String
dataToString dataList =
  List.foldr (++) ""
  <| List.map
        (\elem -> (Tuple.first elem)++" "++
          (let rightString = String.fromFloat <| Maybe.withDefault -1234879231 <| Tuple.second elem
           in if rightString == "-1234879231" then "No Data" else rightString
          )++"\n"
        )
        dataList

-- Für Zeichnen

augmentLevel : List ( Int, Int ) -> List ( ( Int, Int ), ( Int, Int ) )
augmentLevel level =
    let next : List ( Int, Int )
        next = List.foldr
                (\( w, h ) nl ->
                    let ( nlx, nly ) =
                            case nl of
                                []            -> ( 1, 1 )
                                ( x, y ) :: _ -> ( x, y )
                    in ( w * nlx, h * nly ) :: nl
                )
                []
                level
    in List.map2 Tuple.pair level (List.append (List.drop 1 next) [ ( 0, 0 ) ])

coordsListToListOfLines : List (Int, Int) -> List (List Int)
coordsListToListOfLines coords = -- requires the coords to fill the whole place minX - maxX, minY - maxY otherwise there will be -1 elements
    let minX = Maybe.withDefault                    0 <| List.minimum <| List.map Tuple.first  coords
        maxX = Maybe.withDefault (List.length coords) <| List.maximum <| List.map Tuple.first  coords
        minY = Maybe.withDefault                    0 <| List.minimum <| List.map Tuple.second coords
        maxY = Maybe.withDefault (List.length coords) <| List.maximum <| List.map Tuple.second coords
        initialLines = List.repeat (maxY-minY+1) (List.repeat (maxX-minX+1) -1)
        coordsListToListOfLinesHelper : List (Int,(Int, Int)) -> List (List Int) -> List (List Int)
        coordsListToListOfLinesHelper indexedCoords lines =
          case indexedCoords of
            []            -> lines
            (i,(x,y))::cs -> coordsListToListOfLinesHelper cs <| list_set lines (y-minY) (list_set (list_get lines (y-minY)) (x-minX) i)
    in coordsListToListOfLinesHelper (List.indexedMap Tuple.pair coords) initialLines

type alias LoopRec =
    { xOuterLoop : Int, yLoop : Int, outerResult : List ( Int, Int ) }

drawArrangement : Arrangement -> List ( ( Int, Int ), ( Int, Int ) ) -> List ( Int, Int )
drawArrangement arrangement  lvl =
    let drawLineByLine : Int -> Int -> List ( ( Int, Int ), ( Int, Int ) ) -> List ( Int, Int )
        drawLineByLine x y level =
          case level of
            [] -> [ ( x, y ) ]
            ( ( width, height ), ( next_x, next_y ) ) :: reducedLevel ->
              let temp = List.foldl
                           (\_ { xOuterLoop, yLoop, outerResult } ->
                             let innerResult = List.foldl
                                                 (\_ ( xLoop, result ) ->
                                                   ( xLoop + next_x
                                                   , List.append
                                                       result
                                                       (drawLineByLine xLoop yLoop reducedLevel)
                                                   )
                                                 )( xOuterLoop, [] ) <| List.range 1 width
                             in LoopRec xOuterLoop
                                        (yLoop + next_y)
                                        (List.append outerResult (Tuple.second innerResult))
                           )(LoopRec x y []) <| List.range 1 height
              in temp.outerResult
        drawForthLoop : Int -> Int -> List ( ( Int, Int ), ( Int, Int ) ) -> List ( Int, Int )
        drawForthLoop x y level =
          case level of
            [] -> [ ( x, y ) ]
            ( ( width, height ), ( next_x, next_y ) ) :: reducedLevel ->
              let temp = List.foldl
                           (\h { xOuterLoop, yLoop, outerResult } ->
                             let isOdd = modBy 2 h == 0
                                 xOperation     = if isOdd then (-) else (+)
                                 xLoopOperation = if isOdd then (+) ((width-1)*next_x) else (+) 0
                                 innerResult = List.foldl
                                                 (\_ ( xLoop, result ) ->
                                                   ( xOperation xLoop next_x
                                                   , List.append
                                                       result
                                                       (drawForthLoop (xLoopOperation xLoop) yLoop reducedLevel)
                                                   )
                                                 )( xOuterLoop, [] ) <| List.range 1 width
                             in LoopRec xOuterLoop
                                        (yLoop + next_y)
                                        (List.append outerResult (Tuple.second innerResult))
                           )(LoopRec x y []) <| List.range 1 height
              in temp.outerResult
    in case arrangement of
         Line_By_Line -> drawLineByLine 0 0 lvl
         Forth_Loop   -> drawForthLoop  0 0 lvl

drawDimension : ( Int, Int ) -> List ( Int, Int ) -> List ( String, Maybe Float ) -> Svg Msg
drawDimension (w,h) pixelMapping data =
    let scaleTo01 = data
                      |> List.map Tuple.second
                      |> Maybe.Extra.values
                      |> Statistics.extent
                      |> Maybe.withDefault ( 0, 1 )
                      |> Scale.linear ( 0, 1 )
        -- without this multiplier the last col and last row don't fits in viewbox w h
        (minX,maxX) = Maybe.withDefault (0,0) <| Statistics.extent <| List.map toFloat <| List.map Tuple.first  pixelMapping
        (minY,maxY) = Maybe.withDefault (0,0) <| Statistics.extent <| List.map toFloat <| List.map Tuple.second pixelMapping
        w_multiplier = (maxX-minX-1)/(maxX-minX)
        h_multiplier = (maxY-minY-1)/(maxY-minY)

        ( scaleW, scaleH ) =
            pixelMapping
                |> List.map (Tuple.mapBoth toFloat toFloat)
                |> (\l -> ( List.map Tuple.first l, List.map Tuple.second l ))
                |> Tuple.mapBoth
                    (Statistics.extent >> Maybe.withDefault ( 0.0, 100.0 ) >> Scale.linear ( 0, (*) w_multiplier <| toFloat w ))
                    (Statistics.extent >> Maybe.withDefault ( 0.0, 100.0 ) >> Scale.linear ( 0, (*) h_multiplier <| toFloat h ))

        drawPixel : ( Int, Int ) -> ( String, Maybe Float ) -> Svg Msg
        drawPixel ( pixelX, pixelY ) ( label, value ) =
            rect[ TypedSvg.Attributes.x <| TypedSvg.Types.Px <| Scale.convert scaleW <| toFloat pixelX
                , TypedSvg.Attributes.y <| TypedSvg.Types.Px <| Scale.convert scaleH <| toFloat pixelY
                , TypedSvg.Attributes.width <| TypedSvg.Types.Px <| Scale.convert scaleW 1
                , TypedSvg.Attributes.height <| TypedSvg.Types.Px <| Scale.convert scaleH 1
                , TypedSvg.Attributes.fill <| TypedSvg.Types.Paint
                    <| Maybe.withDefault Color.darkGray
                                      <| Maybe.map (Scale.Color.infernoInterpolator << Scale.convert scaleTo01) value
                ][]
    in g [] <| List.map2 drawPixel pixelMapping data

-- Working with Lists

list_Last : List a -> Maybe a
list_Last list =
  case list of
    []    -> Nothing
    x::[] -> Just x
    x::xs -> list_Last xs

list_next : List a -> List a
list_next list = List.append (List.drop 1 list) (List.take 1 list)

list_get : List a -> Int -> a -- returns the listelement at index i. creates endless loop if length list < i+1
list_get list i =
    case list of
        []    -> list_get list i
        x::xs -> if i == 0 then x else list_get xs (i-1)

list_set : List a -> Int -> a -> List a -- resets at index i the element with e. does nothing if length list < i+1
list_set list i e =
    case list of
        []    -> []
        x::xs -> if i==0 then e::xs else x::(list_set xs (i-1) e)

-- Working with Errors
httpErrorToString : Http.Error -> String
httpErrorToString err =
  "Http-Error: "++(
    case err of
      Http.BadUrl string  -> "BadUrl "++string
      Http.Timeout        -> "Timeout"
      Http.NetworkError   -> "NetworkError"
      Http.BadStatus int  -> "BadStatus "++(String.fromInt int)
      Http.BadBody string -> "BadBody "++string
  )